# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""Autocompletion for the editor."""


import re
from typing import Dict, Optional, List, Tuple

from evdev.ecodes import EV_KEY
from gi.repository import Gdk, Gtk, GLib, GObject

from inputremapper.configs.keyboard_layout import keyboard_layout, DISABLE_NAME
from inputremapper.configs.mapping import MappingData
from inputremapper.gui.components.editor import CodeEditor
from inputremapper.gui.controller import Controller
from inputremapper.gui.messages.message_broker import MessageBroker, MessageType
from inputremapper.gui.messages.message_data import UInputsData
from inputremapper.gui.utils import debounce
from inputremapper.injection.macros.parse import Parser
from inputremapper.logging.logger import logger

# no deprecated shorthand function-names
FUNCTION_NAMES = [name for name in Parser.TASK_CLASSES.keys() if len(name) > 1]
# no deprecated functions
FUNCTION_NAMES.remove("ifeq")

Capabilities = Dict[int, List]


def _get_left_text(iter_: Gtk.TextIter) -> str:
    buffer = iter_.get_buffer()
    result = buffer.get_text(buffer.get_start_iter(), iter_, True)
    result = Parser.remove_comments(result)
    result = result.replace("\n", " ")
    return result.lower()


# regex to search for the beginning of a...
PARAMETER = r".*?[(,=+]\s*"
FUNCTION_CHAIN = r".*?\)\s*\.\s*"


def get_incomplete_function_name(iter_: Gtk.TextIter) -> str:
    """Get the word that is written left to the TextIter."""
    left_text = _get_left_text(iter_)

    # match foo in:
    #  bar().foo
    #  bar()\n.foo
    #  bar().\nfoo
    #  bar(\nfoo
    #  bar(\nqux=foo
    #  bar(KEY_A,\nfoo
    #  foo
    match = re.match(rf"(?:{FUNCTION_CHAIN}|{PARAMETER}|^)(\w+)$", left_text)
    logger.debug('get_incomplete_function_name text: "%s" match: %s', left_text, match)

    if match is None:
        return ""

    return match[1]


def get_incomplete_parameter(iter_: Gtk.TextIter) -> Optional[str]:
    """Get the parameter that is written left to the TextIter."""
    left_text = _get_left_text(iter_)

    # match foo in:
    #  bar(foo
    #  bar(a=foo
    #  bar(qux, foo
    #  foo
    #  bar + foo
    match = re.match(rf"(?:{PARAMETER}|^)(\w+)$", left_text)
    logger.debug('get_incomplete_parameter text: "%s" match: %s', left_text, match)

    if match is None:
        return None

    return match[1]


def propose_symbols(text_iter: Gtk.TextIter, codes: List[int]) -> List[Tuple[str, str]]:
    """Find key names that match the input at the cursor and are mapped to the codes."""
    incomplete_name = get_incomplete_parameter(text_iter)

    if incomplete_name is None or len(incomplete_name) <= 1:
        return []

    incomplete_name = incomplete_name.lower()

    names = list(keyboard_layout.list_names(codes=codes)) + [DISABLE_NAME]

    return [
        (name, name)
        for name in names
        if incomplete_name in name.lower() and incomplete_name != name.lower()
    ]


def propose_function_names(text_iter: Gtk.TextIter) -> List[Tuple[str, str]]:
    """Find function names that match the input at the cursor."""
    incomplete_name = get_incomplete_function_name(text_iter)

    if incomplete_name is None or len(incomplete_name) <= 1:
        return []

    incomplete_name = incomplete_name.lower()

    # A list of
    # - ("key", "key(symbol)")
    # - ("repeat", "repeat(repeats, macro)")
    # etc.
    function_names: List[Tuple[str, str]] = []

    for name in FUNCTION_NAMES:
        if incomplete_name in name.lower() and incomplete_name != name.lower():
            task_class = Parser.TASK_CLASSES[name]
            argument_names = task_class.get_macro_argument_names()
            function_names.append((name, f"{name}({', '.join(argument_names)})"))

    return function_names


class SuggestionLabel(Gtk.Label):
    """A label with some extra internal information."""

    __gtype_name__ = "SuggestionLabel"

    def __init__(self, display_name: str, suggestion: str):
        super().__init__(label=display_name)
        self.suggestion = suggestion


class Autocompletion(Gtk.Popover):
    """Provide keyboard-controllable beautiful autocompletions.

    The one provided via source_view.get_completion() is not very appealing
    """

    __gtype_name__ = "Autocompletion"
    _target_uinput: Optional[str] = None

    def __init__(
        self,
        message_broker: MessageBroker,
        controller: Controller,
        code_editor: CodeEditor,
    ):
        """Create an autocompletion popover.

        It will remain hidden until there is something to autocomplete.

        Parameters
        ----------
        code_editor
            The widget that contains the text that should be autocompleted
        """
        super().__init__(
            # Don't switch the focus to the popover when it shows
            modal=False,
            # Always show the popover below the cursor, don't move it to a different
            # position based on the location within the window
            constrain_to=Gtk.PopoverConstraint.NONE,
        )

        self.code_editor = code_editor
        self.controller = controller
        self.message_broker = message_broker
        self._uinputs: Optional[Dict[str, Capabilities]] = None
        self._target_key_capabilities: List[int] = []

        self.scrolled_window = Gtk.ScrolledWindow(
            min_content_width=200,
            max_content_height=200,
            propagate_natural_width=True,
            propagate_natural_height=True,
        )
        self.list_box = Gtk.ListBox()
        self.list_box.get_style_context().add_class("transparent")
        self.scrolled_window.add(self.list_box)

        # row-activated is on-click,
        # row-selected is when scrolling through it
        self.list_box.connect(
            "row-activated",
            self._on_suggestion_clicked,
        )

        self.add(self.scrolled_window)

        self.get_style_context().add_class("autocompletion")

        self.set_position(Gtk.PositionType.BOTTOM)

        self.code_editor.gui.connect("key-press-event", self.navigate)

        # add some delay, so that pressing the button in the completion works before
        # the popover is hidden due to focus-out-event
        self.code_editor.gui.connect("focus-out-event", self.on_gtk_text_input_unfocus)

        self.code_editor.gui.get_buffer().connect("changed", self.update)

        self.set_position(Gtk.PositionType.BOTTOM)

        self.visible = False

        self.attach_to_events()
        self.show_all()
        self.popdown()  # hidden by default. this needs to happen after show_all!

    def attach_to_events(self):
        self.message_broker.subscribe(MessageType.mapping, self._on_mapping_changed)
        self.message_broker.subscribe(MessageType.uinputs, self._on_uinputs_changed)

    def on_gtk_text_input_unfocus(self, *_):
        """The code editor was unfocused."""
        GLib.timeout_add(100, self.popdown)
        # "(input-remapper-gtk:97611): Gtk-WARNING **: 16:33:56.464: GtkTextView -
        # did not receive focus-out-event. If you connect a handler to this signal,
        # it must return FALSE so the text view gets the event as well"
        return False

    def navigate(self, _, event: Gdk.EventKey):
        """Using the keyboard to select an autocompletion suggestion."""
        if not self.visible:
            return

        if event.keyval == Gdk.KEY_Escape:
            self.popdown()
            return

        selected_row = self.list_box.get_selected_row()

        if event.keyval not in [Gdk.KEY_Down, Gdk.KEY_Up, Gdk.KEY_Return]:
            # not one of the keys that controls the autocompletion. Deselect
            # the row but keep it open
            self.list_box.select_row(None)
            return

        if event.keyval == Gdk.KEY_Return:
            if selected_row is None:
                # nothing selected, forward the event to the text editor
                return

            # a row is selected and should be used for autocompletion
            self.list_box.emit("row-activated", selected_row)
            return Gdk.EVENT_STOP

        num_rows = len(self.list_box.get_children())

        if selected_row is None:
            # select the first row
            if event.keyval == Gdk.KEY_Down:
                new_selected_row = self.list_box.get_row_at_index(0)

            if event.keyval == Gdk.KEY_Up:
                new_selected_row = self.list_box.get_row_at_index(num_rows - 1)
        else:
            # select the next row
            selected_index = selected_row.get_index()
            new_index = selected_index

            if event.keyval == Gdk.KEY_Down:
                new_index += 1

            if event.keyval == Gdk.KEY_Up:
                new_index -= 1

            if new_index < 0:
                new_index = num_rows - 1

            if new_index > num_rows - 1:
                new_index = 0

            new_selected_row = self.list_box.get_row_at_index(new_index)

        self.list_box.select_row(new_selected_row)

        self._scroll_to_row(new_selected_row)

        # don't change editor contents
        return Gdk.EVENT_STOP

    def _scroll_to_row(self, row: Gtk.ListBoxRow):
        """Scroll up or down so that the row is visible."""
        # unfortunately, it seems that without focusing the row it won't happen
        # automatically (or whatever the reason for this is, just a wild guess)
        # (the focus should not leave the code editor, so that continuing
        # to write code is possible), so here is a custom solution.
        row_height = row.get_allocation().height

        list_box_height = self.list_box.get_allocated_height()

        if row:
            # get coordinate relative to the list_box,
            # measured from the top of the selected row to the top of the list_box
            row_y_position = row.translate_coordinates(self.list_box, 0, 0)[1]

            # Depending on the theme, the y_offset will be > 0, even though it
            # is the uppermost element, due to margins/paddings.
            if row_y_position < row_height:
                row_y_position = 0

            # if the selected row sits lower than the second to last row,
            # then scroll all the way down. otherwise it will only scroll down
            # to the bottom edge of the selected-row, which might not actually be the
            # bottom of the list-box due to paddings.
            if row_y_position > list_box_height - row_height * 1.5:
                # using a value that is too high doesn't hurt here.
                row_y_position = list_box_height

            # the visible height of the scrolled_window. not the content.
            height = self.scrolled_window.get_max_content_height()

            current_y_scroll = self.scrolled_window.get_vadjustment().get_value()

            vadjustment = self.scrolled_window.get_vadjustment()

            # for the selected row to still be visible, its y_offset has to be
            # at height - row_height. If the y_offset is higher than that, then
            # the autocompletion needs to scroll down to make it visible again.
            if row_y_position > current_y_scroll + (height - row_height):
                value = row_y_position - (height - row_height)
                vadjustment.set_value(value)

            if row_y_position < current_y_scroll:
                # the selected element is not visiable, so we need to scroll up.
                vadjustment.set_value(row_y_position)

    def _get_text_iter_at_cursor(self):
        """Get Gtk.TextIter at the current text cursor location."""
        cursor = self.code_editor.gui.get_cursor_locations()[0]
        return self.code_editor.gui.get_iter_at_location(cursor.x, cursor.y)[1]

    def popup(self):
        self.visible = True
        super().popup()

    def popdown(self):
        self.visible = False
        super().popdown()

    @debounce(100)
    def update(self, *_):
        """Find new autocompletion suggestions and display them. Hide if none."""
        if len(self._target_key_capabilities) == 0:
            logger.error("No target capabilities available")
            return

        if not self.code_editor.gui.is_focus():
            self.popdown()
            return

        self.list_box.forall(self.list_box.remove)

        # move the autocompletion to the text cursor
        cursor = self.code_editor.gui.get_cursor_locations()[0]
        # convert it to window coords, because the cursor values will be very large
        # when the TextView is in a scrolled down ScrolledWindow.
        window_coords = self.code_editor.gui.buffer_to_window_coords(
            Gtk.TextWindowType.TEXT, cursor.x, cursor.y
        )
        cursor.x = window_coords.window_x
        cursor.y = window_coords.window_y
        cursor.y += 12

        if self.code_editor.gui.get_show_line_numbers():
            cursor.x += 48

        self.set_pointing_to(cursor)

        text_iter = self._get_text_iter_at_cursor()
        # get a list of (evdev/xmodmap symbol-name, display-name)
        suggested_names = propose_function_names(text_iter)
        suggested_names += propose_symbols(text_iter, self._target_key_capabilities)

        if len(suggested_names) == 0:
            self.popdown()
            return

        self.popup()  # ffs was this hard to find

        # add visible autocompletion entries
        for suggestion, display_name in suggested_names:
            label = SuggestionLabel(display_name, suggestion)
            self.list_box.insert(label, -1)
            label.show_all()

    def _update_capabilities(self):
        if self._target_uinput and self._uinputs:
            self._target_key_capabilities = self._uinputs[self._target_uinput][EV_KEY]

    def _on_mapping_changed(self, mapping: MappingData):
        self._target_uinput = mapping.target_uinput
        self._update_capabilities()

    def _on_uinputs_changed(self, data: UInputsData):
        self._uinputs = data.uinputs
        self._update_capabilities()

    def _on_suggestion_clicked(self, _, selected_row):
        """An autocompletion suggestion was selected and should be inserted."""
        selected_label = selected_row.get_children()[0]
        suggestion = selected_label.suggestion
        buffer = self.code_editor.gui.get_buffer()

        # make sure to replace the complete unfinished word. Look to the right and
        # remove whatever there is
        cursor_iter = self._get_text_iter_at_cursor()
        right = buffer.get_text(cursor_iter, buffer.get_end_iter(), True)
        match = re.match(r"^(\w+)", right)
        right = match[1] if match else ""
        Gtk.TextView.do_delete_from_cursor(
            self.code_editor.gui, Gtk.DeleteType.CHARS, len(right)
        )

        # do the same to the left
        cursor_iter = self._get_text_iter_at_cursor()
        left = buffer.get_text(buffer.get_start_iter(), cursor_iter, True)
        match = re.match(r".*?(\w+)$", re.sub("\n", " ", left))
        left = match[1] if match else ""
        Gtk.TextView.do_delete_from_cursor(
            self.code_editor.gui, Gtk.DeleteType.CHARS, -len(left)
        )

        # insert the autocompletion
        Gtk.TextView.do_insert_at_cursor(self.code_editor.gui, suggestion)

        self.emit("suggestion-inserted")


GObject.signal_new(
    "suggestion-inserted",
    Autocompletion,
    GObject.SignalFlags.RUN_FIRST,
    None,
    [],
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import asyncio
import unittest

from inputremapper.configs.preset import Preset
from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.context import Context
from inputremapper.injection.global_uinputs import GlobalUInputs, UInput
from inputremapper.injection.macros.macro import Macro, macro_variables
from inputremapper.injection.macros.parse import Parser
from inputremapper.injection.mapping_handlers.mapping_parser import MappingParser
from tests.lib.fixtures import fixtures
from tests.lib.logger import logger
from tests.lib.patches import InputDevice


class MacroTestBase(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls):
        macro_variables.start()

    def setUp(self):
        self.result = []
        self.global_uinputs = GlobalUInputs(UInput)
        self.mapping_parser = MappingParser(self.global_uinputs)

        try:
            self.loop = asyncio.get_event_loop()
        except RuntimeError:
            # suddenly "There is no current event loop in thread 'MainThread'"
            # errors started to appear
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)

        self.source_device = InputDevice(fixtures.bar_device.path)

        self.context = Context(
            Preset(),
            source_devices={fixtures.bar_device.get_device_hash(): self.source_device},
            forward_devices={},
            mapping_parser=self.mapping_parser,
        )

    def tearDown(self):
        self.result = []

    def handler(self, type_: int, code: int, value: int):
        """Where macros should write codes to."""
        logger.info(f"macro wrote{(type_, code, value)}")
        self.result.append((type_, code, value))

    async def trigger_sequence(self, macro: Macro, event):
        for listener in self.context.listeners:
            asyncio.ensure_future(listener(event))
            # this still might cause race conditions and the test to fail
            await asyncio.sleep(0)

        macro.press_trigger()
        if macro.running:
            return
        asyncio.ensure_future(macro.run(self.handler))

    async def release_sequence(self, macro: Macro, event):
        for listener in self.context.listeners:
            asyncio.ensure_future(listener(event))
            # this still might cause race conditions and the test to fail
            await asyncio.sleep(0)

        macro.release_trigger()

    def count_child_macros(self, macro) -> int:
        count = 0
        for task in macro.tasks:
            count += len(task.child_macros)
            for child_macro in task.child_macros:
                count += self.count_child_macros(child_macro)
        return count

    def count_tasks(self, macro) -> int:
        count = len(macro.tasks)
        for task in macro.tasks:
            for child_macro in task.child_macros:
                count += self.count_tasks(child_macro)
        return count

    def expect_string_in_error(self, string: str, macro: str):
        with self.assertRaises(MacroError) as cm:
            Parser.parse(macro, self.context)
        error = str(cm.exception)
        self.assertIn(string, error)


class DummyMapping:
    macro_key_sleep_ms = 10
    rel_rate = 60
    target_uinput = "keyboard + mouse"


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Exceptions that are thrown when configurations are incorrect."""

# can't merge this with exceptions.py, because I want error constructors here to
# be intelligent to avoid redundant code, and they need imports, which would cause
# circular imports.

# pydantic only catches ValueError, TypeError, and AssertionError

from __future__ import annotations

from typing import Optional

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.injection.global_uinputs import GlobalUInputs


class OutputSymbolVariantError(ValueError):
    def __init__(self):
        super().__init__(
            "Missing Argument: Mapping must either contain "
            "`output_symbol` or `output_type` and `output_code`"
        )


class TriggerPointInRangeError(ValueError):
    def __init__(self, input_config):
        super().__init__(
            f"{input_config = } maps an absolute axis to a button, but the "
            "trigger point (event.analog_threshold) is not between -100[%] "
            "and 100[%]"
        )


class OnlyOneAnalogInputError(ValueError):
    def __init__(self, analog_events):
        super().__init__(
            f"Cannot map a combination of multiple analog inputs: {analog_events}"
            "add trigger points (event.value != 0) to map as a button"
        )


class SymbolNotAvailableInTargetError(ValueError):
    def __init__(self, symbol, target):
        code = keyboard_layout.get(symbol)

        fitting_targets = GlobalUInputs.find_fitting_default_uinputs(EV_KEY, code)
        fitting_targets_string = '", "'.join(fitting_targets)

        super().__init__(
            f'The output_symbol "{symbol}" is not available for the "{target}" '
            + f'target. Try "{fitting_targets_string}".'
        )


class OutputSymbolUnknownError(ValueError):
    def __init__(self, symbol: str):
        super().__init__(
            f'The output_symbol "{symbol}" is not a macro and not a valid '
            + "keycode-name"
        )


class MacroButTypeOrCodeSetError(ValueError):
    def __init__(self):
        super().__init__(
            "output_symbol is a macro: output_type " "and output_code must be None"
        )


class SymbolAndCodeMismatchError(ValueError):
    def __init__(self, symbol, code):
        super().__init__(
            "output_symbol and output_code mismatch: "
            f"output macro is {symbol} -> {keyboard_layout.get(symbol)} "
            f"but output_code is {code} -> {keyboard_layout.get_name(code)} "
        )


class WrongMappingTypeForKeyError(ValueError):
    def __init__(self):
        super().__init__("Wrong mapping_type for key input")


class MissingMacroOrKeyError(ValueError):
    def __init__(self):
        super().__init__("Missing macro or key")


class MissingOutputAxisError(ValueError):
    def __init__(self, analog_input_config, output_type):
        super().__init__(
            "Missing output axis: "
            f'"{analog_input_config}" is used as analog input, '
            f"but the {output_type = } is not an axis "
        )


class MacroError(ValueError):
    """Macro syntax errors."""

    def __init__(self, symbol: Optional[str] = None, msg="Error while parsing a macro"):
        self.symbol = symbol
        super().__init__(msg)


def pydantify(error: type):
    """Generate a string as it would appear IN pydantic error types.

    This does not include the base class name, which is transformed to snake case in
    pydantic. Example pydantic error type: "value_error.foobar" for FooBarError.
    """
    # See https://github.com/pydantic/pydantic/discussions/5112
    lower_classname = error.__name__.lower()
    if lower_classname.endswith("error"):
        return lower_classname[: -len("error")]
    return lower_classname



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from tests.lib.tmp import tmp

import logging
import os
import shutil
import unittest

import evdev

from inputremapper.configs.paths import PathUtils
from inputremapper.logging.logger import (
    logger,
    ColorfulFormatter,
)
from tests.lib.test_setup import test_setup


def add_filehandler(log_path: str, debug: bool) -> None:
    """Start logging to a file."""
    log_path = os.path.expanduser(log_path)
    os.makedirs(os.path.dirname(log_path), exist_ok=True)
    file_handler = logging.FileHandler(log_path)
    file_handler.setFormatter(ColorfulFormatter(debug))
    logger.addHandler(file_handler)
    logger.info('Starting logging to "%s"', log_path)


@test_setup
class TestLogger(unittest.TestCase):
    def tearDown(self):
        logger.update_verbosity(debug=True)

        # remove the file handler
        logger.handlers = [
            handler
            for handler in logger.handlers
            if not isinstance(logger.handlers, logging.FileHandler)
        ]
        path = os.path.join(tmp, "logger-test")
        PathUtils.remove(path)

    def test_write(self):
        uinput = evdev.UInput(name="foo")
        path = os.path.join(tmp, "logger-test")
        add_filehandler(path, False)
        logger.write((evdev.ecodes.EV_KEY, evdev.ecodes.KEY_A, 1), uinput)
        with open(path, "r") as f:
            content = f.read()
            self.assertIn(
                'Writing (1, 30, 1) to "foo"',
                content,
            )

    def test_log_info(self):
        logger.update_verbosity(debug=False)
        path = os.path.join(tmp, "logger-test")
        add_filehandler(path, False)
        logger.log_info()
        with open(path, "r") as f:
            content = f.read().lower()
            self.assertIn("input-remapper", content)

    def test_makes_path(self):
        path = os.path.join(tmp, "logger-test")
        if os.path.exists(path):
            shutil.rmtree(path)

        new_path = os.path.join(tmp, "logger-test", "a", "b", "c")
        add_filehandler(new_path, False)
        self.assertTrue(os.path.exists(new_path))

    def test_debug(self):
        path = os.path.join(tmp, "logger-test")
        logger.update_verbosity(True)
        add_filehandler(path, True)
        logger.error("abc")
        logger.warning("foo")
        logger.info("123")
        logger.debug("456")
        logger.debug("789")
        with open(path, "r") as f:
            content = f.read().lower()
            self.assertIn("logger.py", content)

            self.assertIn("error", content)
            self.assertIn("abc", content)

            self.assertIn("warn", content)
            self.assertIn("foo", content)

            self.assertIn("info", content)
            self.assertIn("123", content)

            self.assertIn("debug", content)
            self.assertIn("456", content)

            self.assertIn("debug", content)
            self.assertIn("789", content)

    def test_default(self):
        path = os.path.join(tmp, "logger-test")
        logger.update_verbosity(debug=False)
        add_filehandler(path, False)
        logger.error("abc")
        logger.warning("foo")
        logger.info("123")
        logger.debug("456")
        logger.debug("789")
        with open(path, "r") as f:
            content = f.read().lower()
            self.assertNotIn("logger.py", content)
            self.assertNotIn("line", content)

            self.assertIn("error", content)
            self.assertIn("abc", content)

            self.assertIn("warn", content)
            self.assertIn("foo", content)

            self.assertNotIn("info", content)
            self.assertIn("123", content)

            self.assertNotIn("debug", content)
            self.assertNotIn("456", content)

            self.assertNotIn("debug", content)
            self.assertNotIn("789", content)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from evdev.ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.validation_errors import MacroError
from inputremapper.injection.macros.macro import macro_variables
from inputremapper.injection.macros.parse import Parser
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import MacroTestBase, DummyMapping


@test_setup
class TestSet(MacroTestBase):
    async def test_set_key(self):
        code_b = keyboard_layout.get("b")
        macro = Parser.parse("set(foo, b).key($foo)", self.context, DummyMapping)
        await macro.run(self.handler)
        self.assertListEqual(
            self.result,
            [
                (EV_KEY, code_b, 1),
                (EV_KEY, code_b, 0),
            ],
        )

    async def test_int_is_explicit_string(self):
        await Parser.parse(
            'set( \t"b" \n, "1")',
            self.context,
            DummyMapping,
        ).run(self.handler)
        self.assertEqual(macro_variables.get("b"), "1")

    async def test_int_is_int(self):
        await Parser.parse(
            "set(a, 1)",
            self.context,
            DummyMapping,
        ).run(self.handler)
        self.assertEqual(macro_variables.get("a"), 1)

    async def test_none(self):
        await Parser.parse(
            "set(a, )",
            self.context,
            DummyMapping,
        ).run(self.handler)
        self.assertEqual(macro_variables.get("a"), None)

    async def test_set_case_sensitive_1(self):
        await Parser.parse(
            'set(a, "foo")',
            self.context,
            DummyMapping,
        ).run(self.handler)
        self.assertEqual(macro_variables.get("a"), "foo")
        self.assertEqual(macro_variables.get("A"), None)

    async def test_set_case_sensitive_2(self):
        await Parser.parse(
            'set(A, "foo")',
            self.context,
            DummyMapping,
        ).run(self.handler)
        self.assertEqual(macro_variables.get("A"), "foo")
        self.assertEqual(macro_variables.get("a"), None)

    async def test_raises_error(self):
        self.assertRaises(MacroError, Parser.parse, "set($a, 1)", self.context)
        self.assertRaises(MacroError, Parser.parse, "set(1, 2)", self.context)
        self.assertRaises(MacroError, Parser.parse, "set(+, 2)", self.context)
        self.assertRaises(MacroError, Parser.parse, "set(a(), 2)", self.context)
        self.assertRaises(MacroError, Parser.parse, "set('b,c', 2)", self.context)
        self.assertRaises(MacroError, Parser.parse, 'set("b,c", 2)', self.context)
        Parser.parse("set(A, 2)", self.context)  # no error


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Optional, Any, Union, List, Literal, Type, TYPE_CHECKING

from evdev._ecodes import EV_KEY

from inputremapper.configs.keyboard_layout import keyboard_layout
from inputremapper.configs.validation_errors import (
    MacroError,
    SymbolNotAvailableInTargetError,
)
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.macros.macro import Macro
from inputremapper.injection.macros.variable import Variable

if TYPE_CHECKING:
    from inputremapper.injection.macros.raw_value import RawValue
    from inputremapper.configs.mapping import Mapping


class ArgumentFlags(Enum):
    # No default value is set, and the user has to provide one when using the macro
    required = "required"

    # If used, acts like foo(*bar)
    spread = "spread"


@dataclass
class ArgumentConfig:
    """Definition what kind of arguments a task may take."""

    position: Union[int, Literal[ArgumentFlags.spread]]
    name: str
    types: List[Optional[Type]]
    is_symbol: bool = False
    default: Any = ArgumentFlags.required

    # If True, then the value (which should be a string), is the name of a non-constant
    # variable. Tasks that overwrite their value need this, like `set`. The specified
    # types are those that the current value of that variable may have. For `set` this
    # doesn't matter, but something like `add` requires them to be numbers.
    is_variable_name: bool = False

    def is_required(self) -> bool:
        return self.default == ArgumentFlags.required

    def is_spread(self):
        """Does this Argument store all remaining Variables of a Task as a list?"""
        return self.position == ArgumentFlags.spread


class Argument(ArgumentConfig):
    """Validation of variables and access to their value for Tasks during runtime."""

    _variable: Optional[Variable] = None

    # If the position is set to ArgumentFlags.spread, then _variables will be filled
    # with all remaining positional arguments that were passed to a task.
    _variables: List[Variable]

    _mapping: Optional[Mapping] = None

    def __init__(
        self,
        argument_config: ArgumentConfig,
        mapping: Mapping,
    ) -> None:
        # If a default of None is specified, but None is not an allowed type, then
        # input-remapper has a bug here. Add "None" to your ArgumentConfig.types
        assert not (
            argument_config.default is None and None not in argument_config.types
        )

        self.position = argument_config.position
        self.name = argument_config.name
        self.types = argument_config.types
        self.is_symbol = argument_config.is_symbol
        self.default = argument_config.default
        self.is_variable_name = argument_config.is_variable_name

        self._mapping = mapping
        self._variables = []

    def initialize_variables(self, raw_values: List[RawValue]) -> None:
        """If the macro is supposed to contain multiple variables, set them.
        Should be done during parsing."""
        assert len(self._variables) == 0
        assert self._variable is None
        assert self.is_spread()

        for raw_value in raw_values:
            variable = self._parse_raw_value(raw_value)
            self._variables.append(variable)

    def initialize_variable(self, raw_value: RawValue) -> None:
        """Set the Arguments Variable. Done during parsing."""
        assert len(self._variables) == 0
        assert self._variable is None
        assert not self.is_spread()

        variable = self._parse_raw_value(raw_value)
        self._variable = variable

    def initialize_default(self) -> None:
        """Set the Arguments to its default value. Done during parsing."""
        assert len(self._variables) == 0
        assert self._variable is None
        assert not self.is_spread()

        variable = Variable(value=self.default, const=True)
        self._variable = variable

    def get_value(self) -> Any:
        """To ask for the current value of the variable during runtime."""
        assert not self.is_spread(), f"Use .{self.get_values.__name__}()"
        # If a user passed None as value, it should be a Variable(None, const=True) here.
        # If not, a test or input-remapper is broken.
        assert self._variable is not None

        value = self._variable.get_value()

        if not self._variable.const:
            # Dynamic value. Hasn't been validated yet
            value = self._validate_dynamic_value(self._variable)

        return value

    def get_values(self) -> List[Any]:
        """To ask for the current values of the variables during runtime."""
        assert self.is_spread(), f"Use .{self.get_value.__name__}()"

        values = []
        for variable in self._variables:
            if not variable.const:
                values.append(self._validate_dynamic_value(variable))
            else:
                values.append(variable.get_value())

        return values

    def get_variable_name(self) -> str:
        """If the variable is not const, return its name."""
        assert self._variable is not None
        return self._variable.get_name()

    def contains_macro(self) -> bool:
        """Does the underlying Variable contain another child-macro?"""
        assert self._variable is not None
        return isinstance(self._variable.get_value(), Macro)

    def set_value(self, value: Any) -> Any:
        """To set the value of the underlying Variable during runtime.
        Fails for constants."""
        assert self._variable is not None
        if self._variable.const:
            raise Exception("Can't set value of a constant")

        self._variable.set_value(value)

    def assert_is_symbol(self, symbol: str) -> None:
        """Checks if the key/symbol-name is valid. Like "KEY_A" or "escape".

        Using `is_symbol` on the ArgumentConfig is prefered, which causes it to
        automatically do this for you. But some macros may be a bit more flexible,
        and there we want to assert this ourselves only in certain cases."""
        symbol = str(symbol)
        code = keyboard_layout.get(symbol)

        if code is None:
            raise MacroError(msg=f'Unknown key "{symbol}"')

        if self._mapping is not None:
            target = self._mapping.target_uinput
            if target is not None and not GlobalUInputs.can_default_uinput_emit(
                target, EV_KEY, code
            ):
                raise SymbolNotAvailableInTargetError(symbol, target)

    def _parse_raw_value(self, raw_value: RawValue) -> Variable:
        """Validate and parse."""
        value = raw_value.value

        # The order of steps below matters.

        if isinstance(value, Macro):
            return Variable(value=value, const=True)

        if self.is_variable_name:
            # Treat this as a non-constant variable,
            # even without a `$` in front of its name
            if value.startswith('"'):
                # Remove quotes from the string
                value = value[1:-1]
            return Variable(value=value, const=False)

        if value.startswith("$"):
            # Will be resolved during the macros runtime
            return Variable(value=value[1:], const=False)

        if self.is_symbol:
            if value.startswith('"'):
                value = value[1:-1]
            self.assert_is_symbol(value)
            return Variable(value=value, const=True)

        if (value == "" or value == "None") and None in self.types:
            # I think "" is the deprecated alternative to "None"
            return Variable(value=None, const=True)

        if value.startswith('"') and str in self.types:
            # Something with explicit quotes should never be parsed as a number.
            # Treat it as a string no matter the content.
            value = value[1:-1]
            return Variable(value=value, const=True)

        if float in self.types and "." in value:
            try:
                return Variable(value=float(value), const=True)
            except (ValueError, TypeError):
                pass

        if int in self.types:
            try:
                return Variable(value=int(value), const=True)
            except (ValueError, TypeError):
                pass

        if not value.startswith('"') and ("(" in value or ")" in value):
            # Looks like something that should have been a macro. It is not explicitly
            # wrapped in quotes. Most likely an error. If it was a valid macro, the
            # parser would have parsed it as such.
            raise MacroError(
                msg=f"A broken macro was passed as parameter to {self.name}"
            )

        if str in self.types:
            # Treat as a string. Something like KEY_A in key(KEY_A)
            return Variable(value=value, const=True)

        raise self._type_error_factory(value)

    def _validate_dynamic_value(self, variable: Variable) -> Any:
        """To make sure the value of a non-const variable, asked for at runtime, is
        fitting for the given ArgumentConfig."""
        # Most of the stuff has already been taken care of when, for example,
        # the "1" of set(foo, 1), or the '"bar"' or set(foo, "bar") was parsed the
        # first time. In the first case we get a number 1, and in the second a string
        # `bar` without quotes
        assert not variable.const
        value = variable.get_value()

        if self.is_symbol:
            # value might be int `1`, which is a valid symbol for `key(1)`
            value = str(value)
            self.assert_is_symbol(value)
            return value

        if None in self.types and value is None:
            return value

        if type(value) in self.types:
            return value

        if type(value) not in self.types and str in self.types:
            # `set` cannot make predictions where the variable will be used. Make sure
            # the type is compatible, and turn numbers back into strings if need be.
            return str(value)

        # If the value is "1", we don't attempt to parse it as a number. This being a
        # string means that something like `set(foo, "1")` was used, which enforces a
        # string datatype. Otherwise, `set` would have already turned it into an int.

        raise self._type_error_factory(value)

    def _is_numeric_string(self, value: str) -> bool:
        """Check if the value can be turned into a number."""
        try:
            float(value)
            return True
        except ValueError:
            return False

    def _type_error_factory(self, value: Any) -> MacroError:
        formatted_types: List[str] = []

        for type_ in self.types:
            if type_ is None:
                formatted_types.append("None")
            else:
                formatted_types.append(type_.__name__)

        return MacroError(
            msg=(
                f'Expected "{self.name}" to be one of {formatted_types}, but got '
                f'{type(value).__name__} "{value}"'
            )
        )



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations  # needed for the TYPE_CHECKING import

from typing import TYPE_CHECKING, Dict, Hashable, Tuple

import evdev
from evdev.ecodes import EV_ABS, EV_REL

from inputremapper.configs.input_config import InputCombination
from inputremapper.configs.mapping import Mapping
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    InputEventHandler,
    HandlerEnums,
)
from inputremapper.input_event import InputEvent
from inputremapper.logging.logger import logger

if TYPE_CHECKING:
    from inputremapper.injection.context import Context


class CombinationHandler(MappingHandler):
    """Keeps track of a combination and notifies a sub handler."""

    # map of InputEvent.input_match_hash -> bool , keep track of the combination state
    _pressed_keys: Dict[Hashable, bool]
    # the last update we sent to a sub-handler. If this is true, the output key is
    # still being held down.
    _output_previously_active: bool
    _sub_handler: InputEventHandler
    _handled_input_hashes: list[Hashable]
    _requires_a_release: Dict[Tuple[int, int], bool]

    def __init__(
        self,
        combination: InputCombination,
        mapping: Mapping,
        context: Context,
        global_uinputs: GlobalUInputs,
        **_,
    ) -> None:
        logger.debug(str(mapping))
        super().__init__(combination, mapping, global_uinputs)
        self._pressed_keys = {}
        self._output_previously_active = False
        self._context = context
        self._requires_a_release = {}

        # prepare a key map for all events with non-zero value
        for input_config in combination:
            assert not input_config.defines_analog_input
            self._pressed_keys[input_config.input_match_hash] = False

        self._handled_input_hashes = [
            input_config.input_match_hash for input_config in combination
        ]

        assert len(self._pressed_keys) > 0  # no combination handler without a key

    def __str__(self):
        return (
            f'CombinationHandler for "{str(self.mapping.input_combination)}" '
            f"{tuple(t for t in self._pressed_keys.keys())}"
        )

    def __repr__(self):
        description = (
            f'CombinationHandler for "{repr(self.mapping.input_combination)}" '
            f"{tuple(t for t in self._pressed_keys.keys())}"
        )
        return f"<{description} at {hex(id(self))}>"

    @property
    def child(self):
        # used for logging
        return self._sub_handler

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        if event.input_match_hash not in self._handled_input_hashes:
            # we are not responsible for the event
            return False

        # update the state
        # The value of non-key input should have been changed to either 0 or 1 at this
        # point by other handlers.
        is_pressed = event.value == 1
        self._pressed_keys[event.input_match_hash] = is_pressed
        # maybe this changes the activation status (triggered/not-triggered)
        changed = self._is_activated() != self._output_previously_active

        if changed:
            if is_pressed:
                return self._handle_freshly_activated(suppress, event, source)
            else:
                return self._handle_freshly_deactivated(event, source)
        else:
            if is_pressed:
                return self._handle_no_change_press(event)
            else:
                return self._handle_no_change_release(event)

    def _handle_no_change_press(self, event: InputEvent) -> bool:
        """A key was pressed, but this doesn't change the combinations activation state.
        Can only happen if either the combination wasn't already active, or a duplicate
        key-down event arrived (EV_ABS?)
        """
        # self._output_previously_active is negated, because if the output is active, a
        # key-down event triggered it, which then did not get forwarded, therefore
        # it doesn't require a release.
        self._require_release_later(not self._output_previously_active, event)
        # output is active: consume the event
        # output inactive: forward the event
        return self._output_previously_active

    def _handle_no_change_release(self, event: InputEvent) -> bool:
        """One of the combinations keys was released, but it didn't untrigger the
        combination yet."""
        # Negate: `False` means that the event-reader will forward the release.
        return not self._should_release_event(event)

    def _handle_freshly_activated(
        self,
        suppress: bool,
        event: InputEvent,
        source: evdev.InputDevice,
    ) -> bool:
        """The combination was deactivated, but is activated now."""
        if suppress:
            return False

        # Send key up events to the forwarded uinput if configured to do so.
        self._forward_release()

        logger.debug(
            "Sending %s to sub-handler %s",
            repr(event),
            repr(self._sub_handler),
        )
        self._output_previously_active = bool(event.value)
        sub_handler_result = self._sub_handler.notify(event, source, suppress)

        # Only if the sub-handler return False, we need a release-event later.
        # If it handled the event, the user never sees this key-down event.
        self._require_release_later(not sub_handler_result, event)
        return sub_handler_result

    def _handle_freshly_deactivated(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
    ) -> bool:
        """The combination was activated, but is deactivated now."""
        # We ignore the `suppress` argument for release events. Otherwise, we
        # might end up with stuck keys (test_event_pipeline.test_combination).
        # In the case of output axis, this will enable us to activate multiple
        # axis with the same button.

        logger.debug(
            "Sending %s to sub-handler %s",
            repr(event),
            repr(self._sub_handler),
        )
        self._output_previously_active = bool(event.value)
        self._sub_handler.notify(event, source, suppress=False)

        # Negate: `False` means that the event-reader will forward the release.
        return not self._should_release_event(event)

    def _should_release_event(self, event: InputEvent) -> bool:
        """Check if the key-up event should be forwarded by the event-reader.

        After this, the release event needs to be injected by someone, otherwise the
        dictionary was modified erroneously. If there is no entry, we assume that there
        was no key-down event to release. Maybe a duplicate event arrived.
        """
        # Ensure that all injected key-down events will get their release event
        # injected eventually.
        # If a key-up event arrives that will inactivate the combination, but
        # for which previously a key-down event was injected (because it was
        # an earlier key in the combination chain), then we need to ensure that its
        # release is injected as well. So we get two release events in that case:
        # one for the key, and one for the output.
        assert event.value == 0
        return self._requires_a_release.pop(event.type_and_code, False)

    def _require_release_later(self, require: bool, event: InputEvent) -> None:
        """Remember if this key-down event will need a release event later on."""
        assert event.value == 1
        self._requires_a_release[event.type_and_code] = require

    def reset(self) -> None:
        self._sub_handler.reset()
        for key in self._pressed_keys:
            self._pressed_keys[key] = False
        self._requires_a_release = {}
        self._output_previously_active = False

    def _is_activated(self) -> bool:
        """Return if all keys in the keymap are set to True."""
        return False not in self._pressed_keys.values()

    def _forward_release(self) -> None:
        """Forward a button release for all keys if this is a combination.

        This might cause duplicate key-up events but those are ignored by evdev anyway
        """
        if len(self._pressed_keys) == 1 or not self.mapping.release_combination_keys:
            return

        keys_to_release = filter(
            lambda cfg: self._pressed_keys.get(cfg.input_match_hash),
            self.mapping.input_combination,
        )

        logger.debug("Forwarding release for %s", self.mapping.input_combination)

        for input_config in keys_to_release:
            if not self._requires_a_release.get(input_config.type_and_code):
                continue

            origin_hash = input_config.origin_hash
            if origin_hash is None:
                logger.error(
                    f"Can't forward due to missing origin_hash in {repr(input_config)}"
                )
                continue

            forward_to = self._context.get_forward_uinput(origin_hash)
            logger.write(input_config, forward_to)
            forward_to.write(*input_config.type_and_code, 0)
            forward_to.syn()

            # We are done with this key, forget about it
            del self._requires_a_release[input_config.type_and_code]

    def needs_ranking(self) -> bool:
        return bool(self.input_configs)

    def rank_by(self) -> InputCombination:
        return InputCombination(
            [event for event in self.input_configs if not event.defines_analog_input]
        )

    def wrap_with(self) -> Dict[InputCombination, HandlerEnums]:
        return_dict = {}
        for config in self.input_configs:
            if config.type == EV_ABS and not config.defines_analog_input:
                return_dict[InputCombination([config])] = HandlerEnums.abs2btn

            if config.type == EV_REL and not config.defines_analog_input:
                return_dict[InputCombination([config])] = HandlerEnums.rel2btn

        return return_dict



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


import unittest

from inputremapper.configs.validation_errors import (
    MacroError,
)
from inputremapper.injection.macros.argument import Argument, ArgumentConfig
from inputremapper.injection.macros.macro import Macro, macro_variables
from inputremapper.injection.macros.raw_value import RawValue
from inputremapper.injection.macros.variable import Variable
from tests.lib.test_setup import test_setup
from tests.unit.test_macros.macro_test_base import DummyMapping, MacroTestBase


@test_setup
class TestArgument(MacroTestBase):
    def test_resolve(self):
        self.assertEqual(Variable("a", const=True).get_value(), "a")
        self.assertEqual(Variable(1, const=True).get_value(), 1)
        self.assertEqual(Variable(None, const=True).get_value(), None)

        # $ is part of a custom string here
        self.assertEqual(Variable('"$a"', const=True).get_value(), '"$a"')
        self.assertEqual(Variable("'$a'", const=True).get_value(), "'$a'")
        self.assertEqual(Variable("$a", const=True).get_value(), "$a")

        variable = Variable("a", const=False)
        self.assertEqual(variable.get_value(), None)
        macro_variables["a"] = 1
        self.assertEqual(variable.get_value(), 1)

    def test_type_check(self):
        def test(value, types, name, position):
            argument = Argument(
                ArgumentConfig(
                    types=types,
                    name=name,
                    position=position,
                ),
                DummyMapping(),
            )
            argument.initialize_variable(RawValue(value=value))
            return argument.get_value()

        def test_variable(variable, types, name, position):
            argument = Argument(
                ArgumentConfig(
                    types=types,
                    name=name,
                    position=position,
                ),
                DummyMapping(),
            )
            argument._variable = variable
            return argument.get_value()

        # allows params that can be cast to the target type
        self.assertEqual(test("1", [str, None], "foo", 0), "1")
        self.assertEqual(test("1.2", [str], "foo", 2), "1.2")

        self.assertRaises(
            MacroError,
            lambda: test("1.2", [int], "foo", 3),
        )
        self.assertRaises(MacroError, lambda: test("a", [None], "foo", 0))
        self.assertRaises(MacroError, lambda: test("a", [int], "foo", 1))
        self.assertRaises(
            MacroError,
            lambda: test("a", [int, float], "foo", 2),
        )
        self.assertRaises(
            MacroError,
            lambda: test("a", [int, None], "foo", 3),
        )
        self.assertEqual(test("a", [int, float, None, str], "foo", 4), "a")

        # variables are expected to be of the Variable type here, not a $string
        self.assertRaises(
            MacroError,
            lambda: test("$a", [int], "foo", 4),
        )

        # We don't cast values that were explicitly set as strings back into numbers.
        variable = Variable("a", const=False)
        variable.set_value("5")
        self.assertRaises(
            MacroError,
            lambda: test_variable(variable, [int], "foo", 4),
        )

        self.assertRaises(
            MacroError,
            lambda: test("a", [Macro], "foo", 0),
        )
        self.assertRaises(MacroError, lambda: test("1", [Macro], "foo", 0))

    def test_validate_variable_name(self):
        self.assertRaises(
            MacroError,
            lambda: Variable("1a", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable("$a", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable("a()", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable("1", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable("+", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable("-", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable("*", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable("a,b", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable("a,b", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable("#", const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable(1, const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable(None, const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable([], const=False).validate_variable_name(),
        )
        self.assertRaises(
            MacroError,
            lambda: Variable((), const=False).validate_variable_name(),
        )

        # doesn't raise
        Variable("a", const=False).validate_variable_name()
        Variable("_a", const=False).validate_variable_name()
        Variable("_A", const=False).validate_variable_name()
        Variable("A", const=False).validate_variable_name()
        Variable("Abcd", const=False).validate_variable_name()
        Variable("Abcd_", const=False).validate_variable_name()
        Variable("Abcd_1234", const=False).validate_variable_name()
        Variable("Abcd1234_", const=False).validate_variable_name()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.


"""Share a dictionary across processes."""


import atexit
import multiprocessing
import select
from typing import Optional, Any

from inputremapper.logging.logger import logger


class SharedDict:
    """Share a dictionary across processes."""

    # because unittests terminate all child processes in cleanup I can't use
    # multiprocessing.Manager
    def __init__(self):
        """Create a shared dictionary."""
        super().__init__()

        # To avoid blocking forever if something goes wrong. The maximum
        # observed time communication takes was 0.001 for me on a slow pc
        self._timeout = 0.02

        self.pipe = multiprocessing.Pipe()
        self.process = None
        atexit.register(self._stop)

    def start(self):
        """Ensure the process to manage the dictionary is running."""
        if self.process is not None and self.process.is_alive():
            logger.debug("SharedDict process already running")
            return

        # if the manager has already been running in the past but stopped
        # for some reason, the dictionary contents are lost.
        logger.debug("Starting SharedDict process")
        self.process = multiprocessing.Process(target=self.manage)
        self.process.start()

    def manage(self):
        """Manage the dictionary, handle read and write requests."""
        logger.debug("SharedDict process started")
        shared_dict = {}
        while True:
            message = self.pipe[0].recv()
            logger.debug("SharedDict got %s", message)

            if message[0] == "stop":
                return

            if message[0] == "set":
                shared_dict[message[1]] = message[2]

            if message[0] == "clear":
                shared_dict.clear()

            if message[0] == "get":
                self.pipe[0].send(shared_dict.get(message[1]))

            if message[0] == "ping":
                self.pipe[0].send("pong")

    def _stop(self):
        """Stop the managing process."""
        self.pipe[1].send(("stop",))

    def _clear(self):
        """Clears the memory."""
        self.pipe[1].send(("clear",))

    def get(self, key: str):
        """Get a value from the dictionary.

        If it doesn't exist, returns None.
        """
        return self[key]

    def is_alive(self, timeout: Optional[int] = None):
        """Check if the manager process is running."""
        self.pipe[1].send(("ping",))
        select.select([self.pipe[1]], [], [], timeout or self._timeout)
        if self.pipe[1].poll():
            return self.pipe[1].recv() == "pong"

        return False

    def __setitem__(self, key: str, value: Any):
        self.pipe[1].send(("set", key, value))

    def __getitem__(self, key: str):
        self.pipe[1].send(("get", key))

        select.select([self.pipe[1]], [], [], self._timeout)
        if self.pipe[1].poll():
            return self.pipe[1].recv()

        logger.error("select.select timed out")
        return None

    def __del__(self):
        self._stop()



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from typing import Dict, Tuple, Hashable

import evdev

from inputremapper.configs.input_config import InputCombination
from inputremapper.configs.input_config import InputConfig
from inputremapper.configs.mapping import Mapping
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    HandlerEnums,
    InputEventHandler,
    ContextProtocol,
)
from inputremapper.input_event import InputEvent, EventActions
from inputremapper.logging.logger import logger
from inputremapper.utils import get_device_hash


class AxisSwitchHandler(MappingHandler):
    """Enables or disables an axis.

    Generally, if multiple events are mapped to something in a combination, all of
    them need to be triggered in order to map to the output.

    If an analog input is combined with a key input, then the same thing should happen.
    The key needs to be pressed and the joystick needs to be moved in order to generate
    output.
    """

    _map_axis: InputConfig  # the InputConfig for the axis we switch on or off
    _trigger_keys: Tuple[Hashable, ...]  # all events that can switch the axis
    _active: bool  # whether the axis is on or off
    _last_value: int  # the value of the last axis event that arrived
    _axis_source: evdev.InputDevice  # the cached source of the axis input events
    _forward_device: evdev.UInput  # the cached forward uinput
    _sub_handler: InputEventHandler

    def __init__(
        self,
        combination: InputCombination,
        mapping: Mapping,
        context: ContextProtocol,
        global_uinputs: GlobalUInputs,
        **_,
    ):
        super().__init__(combination, mapping, global_uinputs)
        trigger_keys = tuple(
            event.input_match_hash
            for event in combination
            if not event.defines_analog_input
        )
        assert len(trigger_keys) >= 1
        assert (map_axis := combination.find_analog_input_config())
        self._map_axis = map_axis
        self._trigger_keys = trigger_keys
        self._active = False

        self._last_value = 0
        self._axis_source = None
        self._forward_device = None

        self.context = context

    def __str__(self):
        return f"AxisSwitchHandler for {self._map_axis.type_and_code}"

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def child(self):
        return self._sub_handler

    def _handle_key_input(self, event: InputEvent):
        """If a key is pressed, allow mapping analog events in subhandlers.

        Analog events (e.g. ABS_X, REL_Y) that have gone through Handlers that
        transform them to buttons also count as keys.
        """
        key_is_pressed = bool(event.value)
        if self._active == key_is_pressed:
            # nothing changed
            return False

        self._active = key_is_pressed

        if self._axis_source is None:
            return True

        if not key_is_pressed:
            # recenter the axis
            logger.debug("Stopping axis for %s", self.mapping.input_combination)
            event = InputEvent(
                0,
                0,
                *self._map_axis.type_and_code,
                0,
                actions=(EventActions.recenter,),
                origin_hash=self._map_axis.origin_hash,
            )
            self._sub_handler.notify(event, self._axis_source)
            return True

        if self._map_axis.type == evdev.ecodes.EV_ABS:
            # send the last cached value so that the abs axis
            # is at the correct position
            logger.debug("Starting axis for %s", self.mapping.input_combination)
            event = InputEvent(
                0,
                0,
                *self._map_axis.type_and_code,
                self._last_value,
                origin_hash=self._map_axis.origin_hash,
            )
            self._sub_handler.notify(event, self._axis_source)
            return True

        return True

    def _should_map(self, event: InputEvent):
        return (
            event.input_match_hash in self._trigger_keys
            or event.input_match_hash == self._map_axis.input_match_hash
        )

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        if not self._should_map(event):
            return False

        if event.is_key_event:
            return self._handle_key_input(event)

        # do some caching so that we can generate the
        # recenter event and an initial abs event
        if self._axis_source is None:
            self._axis_source = source

        if self._forward_device is None:
            device_hash = get_device_hash(source)
            self._forward_device = self.context.get_forward_uinput(device_hash)

        # always cache the value
        self._last_value = event.value

        if self._active:
            return self._sub_handler.notify(event, source, suppress)

        return False

    def reset(self) -> None:
        self._last_value = 0
        self._active = False
        self._sub_handler.reset()

    def needs_wrapping(self) -> bool:
        return True

    def wrap_with(self) -> Dict[InputCombination, HandlerEnums]:
        combination = [
            config for config in self.input_configs if not config.defines_analog_input
        ]
        return {InputCombination(combination): HandlerEnums.combination}



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

"""Find, classify and group devices.

Because usually connected devices pop up multiple times in /dev/input,
in order to provide multiple types of input devices (e.g. a keyboard and a
graphics-tablet at the same time)

Those groups are what is being displayed in the device dropdown, and
events are being read from all of the paths of an individual group in the gui
and the injector.
"""

from __future__ import annotations

import asyncio
import enum
import json
import multiprocessing
import os
import re
import threading
from typing import List, Optional

import evdev
from evdev.ecodes import (
    EV_KEY,
    EV_ABS,
    KEY_CAMERA,
    EV_REL,
    BTN_STYLUS,
    ABS_MT_POSITION_X,
    REL_X,
    KEY_A,
    BTN_LEFT,
    REL_Y,
    REL_WHEEL,
)

from inputremapper.configs.paths import PathUtils
from inputremapper.logging.logger import logger
from inputremapper.utils import get_device_hash

TABLET_KEYS = [
    evdev.ecodes.BTN_STYLUS,
    evdev.ecodes.BTN_TOOL_BRUSH,
    evdev.ecodes.BTN_TOOL_PEN,
    evdev.ecodes.BTN_TOOL_RUBBER,
]


class DeviceType(str, enum.Enum):
    GAMEPAD = "gamepad"
    KEYBOARD = "keyboard"
    MOUSE = "mouse"
    TOUCHPAD = "touchpad"
    GRAPHICS_TABLET = "graphics-tablet"
    CAMERA = "camera"
    UNKNOWN = "unknown"


if not hasattr(evdev.InputDevice, "path"):
    # for evdev < 1.0.0 patch the path property
    @property
    def path(device):
        return device.fn

    evdev.InputDevice.path = path


def _is_gamepad(capabilities):
    """Check if joystick movements are available for preset."""
    # A few buttons that indicate a gamepad
    buttons = {
        evdev.ecodes.BTN_BASE,
        evdev.ecodes.BTN_A,
        evdev.ecodes.BTN_THUMB,
        evdev.ecodes.BTN_TOP,
        evdev.ecodes.BTN_DPAD_DOWN,
        evdev.ecodes.BTN_GAMEPAD,
    }
    if not buttons.intersection(capabilities.get(EV_KEY, [])):
        # no button is in the key capabilities
        return False

    # joysticks
    abs_capabilities = capabilities.get(EV_ABS, [])
    if evdev.ecodes.ABS_X not in abs_capabilities:
        return False
    if evdev.ecodes.ABS_Y not in abs_capabilities:
        return False

    return True


def _is_mouse(capabilities):
    """Check if the capabilities represent those of a mouse."""
    # Based on observation, those capabilities need to be present to get an
    # UInput recognized as mouse

    # mouse movements
    if REL_X not in capabilities.get(EV_REL, []):
        return False
    if REL_Y not in capabilities.get(EV_REL, []):
        return False

    # at least the vertical mouse wheel
    if REL_WHEEL not in capabilities.get(EV_REL, []):
        return False

    # and a mouse click button
    if BTN_LEFT not in capabilities.get(EV_KEY, []):
        return False

    return True


def _is_graphics_tablet(capabilities):
    """Check if the capabilities represent those of a graphics tablet."""
    if BTN_STYLUS in capabilities.get(EV_KEY, []):
        return True
    return False


def _is_touchpad(capabilities):
    """Check if the capabilities represent those of a touchpad."""
    if ABS_MT_POSITION_X in capabilities.get(EV_ABS, []):
        return True
    return False


def _is_keyboard(capabilities):
    """Check if the capabilities represent those of a keyboard."""
    if KEY_A in capabilities.get(EV_KEY, []):
        return True
    return False


def _is_camera(capabilities):
    """Check if the capabilities represent those of a camera."""
    key_capa = capabilities.get(EV_KEY)
    return key_capa and len(key_capa) == 1 and key_capa[0] == KEY_CAMERA


def classify(device) -> DeviceType:
    """Figure out what kind of device this is.

    Use this instead of functions like _is_keyboard to avoid getting false
    positives.
    """
    capabilities = device.capabilities(absinfo=False)

    if _is_graphics_tablet(capabilities):
        # check this before is_gamepad to avoid classifying abs_x
        # as joysticks when they are actually stylus positions
        return DeviceType.GRAPHICS_TABLET

    if _is_touchpad(capabilities):
        return DeviceType.TOUCHPAD

    if _is_gamepad(capabilities):
        return DeviceType.GAMEPAD

    if _is_mouse(capabilities):
        return DeviceType.MOUSE

    if _is_camera(capabilities):
        return DeviceType.CAMERA

    if _is_keyboard(capabilities):
        # very low in the chain to avoid classifying most devices
        # as keyboard, because there are many with ev_key capabilities
        return DeviceType.KEYBOARD

    return DeviceType.UNKNOWN


DENYLIST = [".*Yubico.*YubiKey.*", "Eee PC WMI hotkeys"]


def is_denylisted(device: evdev.InputDevice):
    """Check if a device should not be used in input-remapper.

    Parameters
    ----------
    device
    """
    for name in DENYLIST:
        if re.match(name, str(device.name), re.IGNORECASE):
            return True

    return False


def get_unique_key(device: evdev.InputDevice):
    """Find a string key that is unique for a single hardware device.

    All InputDevices in /dev/input that originate from the same physical
    hardware device should return the same key via this function.
    """
    # Keys that should not be used:
    # - device.phys is empty sometimes and varies across virtual
    #   subdevices
    # - device.version varies across subdevices
    return (
        # device.info bustype, vendor and product are unique for
        # a product, but multiple similar device models would be grouped
        # in the same group
        f"{device.info.bustype}_"
        f"{device.info.vendor}_"
        f"{device.info.product}_"
        # device.uniq is empty most of the time. It seems to be the only way to
        # distinguish multiple connected bluetooth gamepads
        f"{device.uniq}_"
        # deivce.phys if "/input..." is removed from it, because the first
        # chunk seems to be unique per hardware (if it's not completely empty)
        f'{device.phys.split("/")[0] or "-"}'
    )


class _Group:
    """Groups multiple devnodes together.

    For example, name could be "Logitech USB Keyboard", devices
    might contain "Logitech USB Keyboard System Control" and "Logitech USB
    Keyboard". paths is a list of files in /dev/input that belong to the
    devices.

    They are grouped by usb port.

    Members
    -------
    name : str
        A human readable name, generated from .names, that should always
        look the same for a device model. It is used to generate the
        presets folder structure
    """

    def __init__(
        self,
        paths: List[os.PathLike],
        names: List[str],
        types: List[DeviceType | str],
        key: str,
    ):
        """Specify a group

        Parameters
        ----------
        paths
            Paths in /dev/input of the grouped devices
        names
            Names of the grouped devices
        types
            Types of the grouped devices
        key
            Unique identifier of the group.

            It should be human readable and if possible equal to group.name.
            To avoid multiple groups having the same key, a number starting
            with 2 followed by a whitespace should be added to it:
            "key", "key 2", "key 3", ...

            This is important for the autoloading configuration. If the key
            changed over reboots, then autoloading would break.
        """
        # There might be multiple groups with the same name here when two
        # similar devices are connected to the computer.
        self.name: str = sorted(names, key=len)[0]

        self.key = key

        self.paths = paths
        self.names = names
        self.types = [DeviceType(type_) for type_ in types]

    def get_preset_path(self, preset: Optional[str] = None):
        """Get a path to the stored preset, or to store a preset to.

        This path is unique per device-model, not per group. Groups
        of the same model share the same preset paths.
        """
        return PathUtils.get_preset_path(self.name, preset)

    def get_devices(self) -> List[evdev.InputDevice]:
        devices: List[evdev.InputDevice] = []
        for path in self.paths:
            try:
                devices.append(evdev.InputDevice(path))
            except (FileNotFoundError, OSError):
                logger.error('Could not find "%s"', path)
                continue
        return devices

    def dumps(self):
        """Return a string representing this object."""
        return json.dumps(
            dict(paths=self.paths, names=self.names, types=self.types, key=self.key),
        )

    @classmethod
    def loads(cls, serialized: str):
        """Load a serialized representation."""
        group = cls(**json.loads(serialized))
        return group

    def __repr__(self):
        return f"<Group ({self.key}) at {hex(id(self))}>"


class _FindGroups(threading.Thread):
    """Thread to get the devices that can be worked with.

    Since InputDevice destructors take quite some time, do this
    asynchronously so that they can take as much time as they want without
    slowing down the initialization.
    """

    def __init__(self, pipe: multiprocessing.Pipe):
        """Construct the process.

        Parameters
        ----------
        pipe
            used to communicate the result
        """
        self.pipe = pipe
        super().__init__()

    def run(self):
        """Do what get_groups describes."""
        # evdev needs asyncio to work
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        logger.debug("Discovering device paths")

        # group them together by usb device because there could be stuff like
        # "Logitech USB Keyboard" and "Logitech USB Keyboard Consumer Control"
        grouped = {}
        for path in evdev.list_devices():
            try:
                device = evdev.InputDevice(path)
            except Exception as error:
                # Observed exceptions in journalctl:
                # - "SystemError: <built-in function ioctl_EVIOCGVERSION> returned NULL
                # without setting an error"
                # - "FileNotFoundError: [Errno 2] No such file or directory:
                # '/dev/input/event12'"
                logger.error(
                    'Failed to access path "%s": %s %s',
                    path,
                    error.__class__.__name__,
                    str(error),
                )
                continue

            if device.name == "Power Button":
                continue

            device_type = classify(device)

            if device_type == DeviceType.CAMERA:
                continue

            # https://www.kernel.org/doc/html/latest/input/event-codes.html
            capabilities = device.capabilities(absinfo=False)

            key_capa = capabilities.get(EV_KEY)
            abs_capa = capabilities.get(EV_ABS)
            rel_capa = capabilities.get(EV_REL)

            if key_capa is None and abs_capa is None and rel_capa is None:
                # skip devices that don't provide buttons or axes that can be mapped
                logger.debug('"%s" has no useful capabilities', device.name)
                continue

            if is_denylisted(device):
                logger.debug('"%s" is denylisted', device.name)
                continue

            key = get_unique_key(device)
            if grouped.get(key) is None:
                grouped[key] = []

            logger.debug(
                'Found %s "%s" at "%s", hash "%s", key "%s"',
                device_type.value,
                device.name,
                path,
                get_device_hash(device),
                key,
            )

            grouped[key].append((device.name, path, device_type))

        # now write down all the paths of that group
        result = []
        used_keys = set()
        for group in grouped.values():
            names = [entry[0] for entry in group]
            devs = [entry[1] for entry in group]

            # generate a human readable key
            shortest_name = sorted(names, key=len)[0]
            key = shortest_name
            i = 2
            while key in used_keys:
                key = f"{shortest_name} {i}"
                i += 1
            used_keys.add(key)

            group = _Group(
                key=key,
                paths=devs,
                names=names,
                types=sorted(
                    list({item[2] for item in group if item[2] != DeviceType.UNKNOWN})
                ),
            )

            result.append(group.dumps())

        self.pipe.send(json.dumps(result))
        loop.close()  # avoid resource allocation warnings
        # now that everything is sent via the pipe, the InputDevice
        # destructors can go on and take ages to complete in the thread
        # without blocking anything


class _Groups:
    """Contains and manages all groups."""

    def __init__(self):
        self._groups: List[_Group] = None

    def __getattribute__(self, key: str):
        """To lazy load group info only when needed.

        For example, this helps to keep logs of input-remapper-control clear when it
        doesn't need it the information.
        """
        if key == "_groups" and object.__getattribute__(self, "_groups") is None:
            object.__setattr__(self, "_groups", [])
            object.__getattribute__(self, "refresh")()

        return object.__getattribute__(self, key)

    def refresh(self):
        """Look for devices and group them together.

        Since this needs to do some stuff with /dev and spawn processes the
        result is cached. Use refresh_groups if you need up to date
        devices.
        """
        pipe = multiprocessing.Pipe()
        _FindGroups(pipe[1]).start()
        # block until groups are available
        self.loads(pipe[0].recv())

        if len(self._groups) == 0:
            logger.error("Did not find any input device")
        else:
            keys = [f'"{group.key}"' for group in self._groups]
            logger.info("Found %s", ", ".join(keys))

    def filter(self, include_inputremapper: bool = False) -> List[_Group]:
        """Filter groups."""
        result = []
        for group in self._groups:
            name = group.name
            if not include_inputremapper and name.startswith("input-remapper"):
                continue

            result.append(group)

        return result

    def set_groups(self, new_groups: List[_Group]):
        """Overwrite all groups."""
        logger.debug("Overwriting groups with %s", new_groups)
        self._groups = new_groups

    def list_group_names(self) -> List[str]:
        """Return a list of all 'name' properties of the groups."""
        return [
            group.name
            for group in self._groups
            if not group.name.startswith("input-remapper")
        ]

    def __len__(self):
        return len(self._groups)

    def __iter__(self):
        return iter(self._groups)

    def dumps(self):
        """Create a deserializable string representation."""
        return json.dumps([group.dumps() for group in self._groups])

    def loads(self, dump: str):
        """Load a serialized representation created via dumps."""
        self._groups = [_Group.loads(group) for group in json.loads(dump)]

    def find(
        self,
        name: Optional[str] = None,
        key: Optional[str] = None,
        path: Optional[str] = None,
        include_inputremapper: bool = False,
    ) -> Optional[_Group]:
        """Find a group that matches the provided parameters.

        Parameters
        ----------
        name
            "USB Keyboard"
            Not unique, will return the first group that matches.
        key
            "USB Keyboard", "USB Keyboard 2", ...
        path
            "/dev/input/event3"
        """
        for group in self._groups:
            if not include_inputremapper and group.name.startswith("input-remapper"):
                continue

            if name and group.name != name:
                continue

            if key and group.key != key:
                continue

            if path and path not in group.paths:
                continue

            return group

        return None


# TODO global objects are bad practice
groups = _Groups()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

import unittest

import evdev
from dataclasses import FrozenInstanceError
from inputremapper.input_event import InputEvent
from tests.lib.test_setup import test_setup


@test_setup
class TestInputEvent(unittest.TestCase):
    def test_from_event(self):
        e1 = InputEvent.from_event(evdev.InputEvent(1, 2, 3, 4, 5))
        e2 = InputEvent.from_event(e1)

        self.assertEqual(e1, e2)
        self.assertEqual(e1.sec, 1)
        self.assertEqual(e1.usec, 2)
        self.assertEqual(e1.type, 3)
        self.assertEqual(e1.code, 4)
        self.assertEqual(e1.value, 5)

        self.assertEqual(e1.sec, e2.sec)
        self.assertEqual(e1.usec, e2.usec)
        self.assertEqual(e1.type, e2.type)
        self.assertEqual(e1.code, e2.code)
        self.assertEqual(e1.value, e2.value)

        self.assertRaises(TypeError, InputEvent.from_event, "1,2,3")

    def test_from_event_tuple(self):
        t1 = (1, 2, 3)
        t2 = (1, "2", 3)
        t3 = (1, 2, 3, 4, 5)
        t4 = (1, "b", 3)

        e1 = InputEvent.from_tuple(t1)
        e2 = InputEvent.from_tuple(t2)
        self.assertEqual(e1, e2)
        self.assertEqual(e1.sec, 0)
        self.assertEqual(e1.usec, 0)
        self.assertEqual(e1.type, 1)
        self.assertEqual(e1.code, 2)
        self.assertEqual(e1.value, 3)

    def test_properties(self):
        e1 = InputEvent.from_tuple((evdev.ecodes.EV_KEY, evdev.ecodes.BTN_LEFT, 1))
        self.assertEqual(
            e1.event_tuple,
            (evdev.ecodes.EV_KEY, evdev.ecodes.BTN_LEFT, 1),
        )
        self.assertEqual(e1.type_and_code, (evdev.ecodes.EV_KEY, evdev.ecodes.BTN_LEFT))

        with self.assertRaises(
            FrozenInstanceError
        ):  # would be TypeError on a slotted class
            e1.event_tuple = (1, 2, 3)

        with self.assertRaises(
            FrozenInstanceError
        ):  # would be TypeError on a slotted class
            e1.type_and_code = (1, 2)

        with self.assertRaises(FrozenInstanceError):
            e1.value = 5

    def test_modify(self):
        e1 = InputEvent(1, 2, 3, 4, 5)
        e2 = e1.modify(value=6)
        e3 = e1.modify(sec=0, usec=0, type_=0, code=0, value=0)

        self.assertNotEqual(e1, e2)
        self.assertEqual(e1.sec, e2.sec)
        self.assertEqual(e1.usec, e2.usec)
        self.assertEqual(e1.type, e2.type)
        self.assertEqual(e1.code, e2.code)
        self.assertNotEqual(e1.value, e2.value)
        self.assertEqual(e3.sec, 0)
        self.assertEqual(e3.usec, 0)
        self.assertEqual(e3.type, 0)
        self.assertEqual(e3.code, 0)
        self.assertEqual(e3.value, 0)

    def test_is_wheel_event(self):
        input_event_x = InputEvent(
            0,
            0,
            evdev.ecodes.EV_REL,
            evdev.ecodes.REL_X,
            1,
        )
        self.assertFalse(input_event_x.is_wheel_event)
        self.assertFalse(input_event_x.is_wheel_hi_res_event)

        input_event_wheel = InputEvent(
            0,
            0,
            evdev.ecodes.EV_REL,
            evdev.ecodes.REL_WHEEL,
            1,
        )
        self.assertTrue(input_event_wheel.is_wheel_event)
        self.assertFalse(input_event_wheel.is_wheel_hi_res_event)

        input_event_wheel_hi_res = InputEvent(
            0,
            0,
            evdev.ecodes.EV_REL,
            evdev.ecodes.REL_WHEEL_HI_RES,
            1,
        )
        self.assertFalse(input_event_wheel_hi_res.is_wheel_event)
        self.assertTrue(input_event_wheel_hi_res.is_wheel_hi_res_event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from evdev.ecodes import (
    LED_NUML,
    LED_CAPSL,
)

from inputremapper.injection.macros.argument import ArgumentConfig
from inputremapper.injection.macros.macro import Macro
from inputremapper.injection.macros.task import Task


class IfLedTask(Task):
    argument_configs = [
        ArgumentConfig(
            name="then",
            position=0,
            types=[Macro, None],
            default=None,
        ),
        ArgumentConfig(
            name="else",
            position=1,
            types=[Macro, None],
            default=None,
        ),
    ]

    led_code = None

    async def run(self, callback) -> None:
        then = self.get_argument("then").get_value()
        else_ = self.get_argument("else").get_value()

        # self.context is only None when the frontend is parsing the macro
        assert self.context is not None
        led_on = self.led_code in self.context.get_leds()

        if led_on:
            if then is not None:
                await then.run(callback)
        elif else_ is not None:
            await else_.run(callback)


class IfNumlockTask(IfLedTask):
    led_code = LED_NUML


class IfCapslockTask(IfLedTask):
    led_code = LED_CAPSL



# -*- coding: utf-8 -*-
# input-remapper - GUI for device specific keyboard mappings
# Copyright (C) 2025 sezanzeb <b8x45ygc9@mozmail.com>
#
# This file is part of input-remapper.
#
# input-remapper is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# input-remapper is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with input-remapper.  If not, see <https://www.gnu.org/licenses/>.

from typing import Tuple, Optional, Dict

import evdev
from evdev.ecodes import EV_ABS

from inputremapper import exceptions
from inputremapper.configs.input_config import InputCombination, InputConfig
from inputremapper.configs.mapping import Mapping
from inputremapper.injection.global_uinputs import GlobalUInputs
from inputremapper.injection.mapping_handlers.axis_transform import Transformation
from inputremapper.injection.mapping_handlers.mapping_handler import (
    MappingHandler,
    HandlerEnums,
    InputEventHandler,
)
from inputremapper.input_event import InputEvent, EventActions
from inputremapper.logging.logger import logger
from inputremapper.utils import get_evdev_constant_name


class AbsToAbsHandler(MappingHandler):
    """Handler which transforms EV_ABS to EV_ABS events."""

    _map_axis: InputConfig  # the InputConfig for the axis we map
    _output_axis: Tuple[int, int]  # the (type, code) of the output axis
    _transform: Optional[Transformation]
    _target_absinfo: evdev.AbsInfo

    def __init__(
        self,
        combination: InputCombination,
        mapping: Mapping,
        global_uinputs: GlobalUInputs,
        **_,
    ) -> None:
        super().__init__(combination, mapping, global_uinputs)

        # find the input event we are supposed to map. If the input combination is
        # BTN_A + ABS_X + BTN_B, then use the value of ABS_X for the transformation
        assert (map_axis := combination.find_analog_input_config(type_=EV_ABS))
        self._map_axis = map_axis

        assert mapping.output_code is not None
        assert mapping.output_type == EV_ABS
        self._output_axis = (mapping.output_type, mapping.output_code)

        target_uinput = global_uinputs.get_uinput(mapping.target_uinput)
        assert target_uinput is not None
        abs_capabilities = target_uinput.capabilities(absinfo=True)[EV_ABS]
        self._target_absinfo = dict(abs_capabilities)[mapping.output_code]

        self._transform = None

    def __str__(self):
        name = get_evdev_constant_name(*self._map_axis.type_and_code)
        return f'AbsToAbsHandler for "{name}" {self._map_axis}'

    def __repr__(self):
        return f"<{str(self)} at {hex(id(self))}>"

    @property
    def child(self):  # used for logging
        return (
            f"maps to: {self.mapping.get_output_name_constant()} "
            f"{self.mapping.get_output_type_code()} at "
            f"{self.mapping.target_uinput}"
        )

    def notify(
        self,
        event: InputEvent,
        source: evdev.InputDevice,
        suppress: bool = False,
    ) -> bool:
        if event.input_match_hash != self._map_axis.input_match_hash:
            return False

        if EventActions.recenter in event.actions:
            self._write(self._scale_to_target(0))
            return True

        if not self._transform:
            absinfo = dict(source.capabilities(absinfo=True)[EV_ABS])[event.code]
            self._transform = Transformation(
                max_=absinfo.max,
                min_=absinfo.min,
                deadzone=self.mapping.deadzone,
                gain=self.mapping.gain,
                expo=self.mapping.expo,
            )

        try:
            self._write(self._scale_to_target(self._transform(event.value)))
            return True
        except (exceptions.UinputNotAvailable, exceptions.EventNotHandled):
            return False

    def reset(self) -> None:
        self._write(self._scale_to_target(0))

    def _scale_to_target(self, x: float) -> int:
        """Scales a x value between -1 and 1 to an integer between
        target_absinfo.min and target_absinfo.max

        input values above 1 or below -1 are clamped to the extreme values
        """
        factor = (self._target_absinfo.max - self._target_absinfo.min) / 2
        offset = self._target_absinfo.min + factor
        y = factor * x + offset
        if y > offset:
            return int(min(self._target_absinfo.max, y))
        else:
            return int(max(self._target_absinfo.min, y))

    def _write(self, value: int):
        """Inject."""
        try:
            self.global_uinputs.write(
                (*self._output_axis, value), self.mapping.target_uinput
            )
        except OverflowError:
            # screwed up the calculation of the event value
            logger.error("OverflowError (%s, %s, %s)", *self._output_axis, value)

    def needs_wrapping(self) -> bool:
        return len(self.input_configs) > 1

    def set_sub_handler(self, handler: InputEventHandler) -> None:
        assert False  # cannot have a sub-handler

    def wrap_with(self) -> Dict[InputCombination, HandlerEnums]:
        if self.needs_wrapping():
            return {InputCombination(self.input_configs): HandlerEnums.axisswitch}
        return {}
